#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
using System.Reflection;
using Extensions;
#endregion

namespace Cotyounoyume{
	public class Orgasm
	{
		#region common parameters
		public Atom atom;
		public EBATCore core;
		
		public bool disableLog = true;
		public float orgasmOpeningTime;
		public float orgasmUpdateTime;
		public int orgasmState = -1;

		public int orgasmAfterglowCount;
		public float orgasmAfterglowIntervalPlus = 0.3f;
		public bool isBigWave = false;
		public bool oncePerCycle = false;
		public int firstTermCountThreshold = 2;
		public int firstTermCount = 0;
		public float orgasmStartInterval = 0f;
		public float orgasmTimeOffset1 = 0.05f;
		public float orgasmTimeOffset2 = 0.05f;
		public float orgasmTimeRandomOffset1;
		public float orgasmTimeRandomOffset2;
		public float orgasmTimeStart = 0f;
		public float orgasmTimePeak = 0f;
		public float orgasmTimeFinish = 0f;
		
		public float continuousBigWaveRatio = 0.1f;
		public float orgasmCooldownTime;
		public float orgasmCooldownTimeOffset = 6.0f;
		
		// commonWaveCycleInterval, commonWaveCycleSmall, commonWaveCycleBig, commonStrength, 
		public float commonWaveCycleInterval;
		public float commonStrength;
		public int commonWaveCycleBig;
		public int commonWaveCycleSmall;

		#endregion
		
		public Orgasm(EBATCore core){
			atom = core.atom;
			this.core = core;
		}
		
		public void LoadSettings(){
			try{
				#region general core.settings from json
				disableLog = core.disableLog;
				core.orgasmStrengthMax      = core.settings["Orgasm"]["Orgasm Strength Max"].AsFloat;
				core.afterglowStrengthMax   = core.settings["Orgasm"]["Afterglow Strength Max"].AsFloat;
				core.headController.bendBackwardThreshold   = core.settings["Orgasm"]["Bending Backward Ratio"].AsFloat;
				
				core.orgasmWaveCycleBigMax   = core.settings["Orgasm"]["Orgasm Wave Big Cycle Count Max"].AsInt;
				core.orgasmWaveCycleBigMin   = core.settings["Orgasm"]["Orgasm Wave Big Cycle Count Min"].AsInt;
				core.orgasmWaveCycleSmallMax = core.settings["Orgasm"]["Orgasm Wave Small Cycle Count Max"].AsInt;
				core.orgasmWaveCycleSmallMin = core.settings["Orgasm"]["Orgasm Wave Small Cycle Count Min"].AsInt;
				core.orgasmAfterglowCycleBigMax   = core.settings["Orgasm"]["Orgasm Afterglow Big Cycle Count Max"].AsInt;
				core.orgasmAfterglowCycleBigMin   = core.settings["Orgasm"]["Orgasm Afterglow Big Cycle Count Min"].AsInt;
				core.orgasmAfterglowCycleSmallMax = core.settings["Orgasm"]["Orgasm Afterglow Small Cycle Count Max"].AsInt;
				core.orgasmAfterglowCycleSmallMin = core.settings["Orgasm"]["Orgasm Afterglow Small Cycle Count Min"].AsInt;
				core.orgasmWaveCycleIntervalBigMax        = core.settings["Orgasm"]["Orgasm Wave Big Cycle Interval Max"].AsFloat;
				core.orgasmWaveCycleIntervalSmallMax      = core.settings["Orgasm"]["Orgasm Wave Small Cycle Interval Max"].AsFloat;
				core.orgasmAfterglowCycleIntervalBigMax   = core.settings["Orgasm"]["Orgasm Afterglow Big Cycle Interval Max"].AsFloat;
				core.orgasmAfterglowCycleIntervalSmallMax = core.settings["Orgasm"]["Orgasm Afterglow Big Small Interval Max"].AsFloat;
				#endregion
				
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void ApplySettings(){
			try{

			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void Maintenance(){
			try{

			}catch (Exception e){
				core.LoggerE(e);
			}
		}

		public void OrgasmCheck(float currentTime){
			try{
				if(core.GetGauge() >= core.stageThreshold[5]){
					if(core.isDuringOrgasm == false || (core.isDuringOrgasm == true && core.isDuringAfterglow == true)){
						if(currentTime > orgasmCooldownTime){
							OrgasmStart();
						}
					}
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}

		public void Update() { }
		
		public void Manage() { }

		public void MakeHerOrgasm()
		{
			if (core.GetGauge() < core.stageThreshold[5])
			{
				core.SetGauge(core.stageThreshold[5]);
			}
			core.UpdateAll();
			core.OnErogenousZoneTouch("VaginaDeeper");
		}

		public void StopHerOrgasm()
		{
			if (core.GetGauge() >= core.stageThreshold[5])
			{
				core.SetGauge(core.stageThreshold[5] - 1);
			}
			OrgasmFinish();
			core.UpdateAll();
		}

		public void OrgasmStart(){
			try{
				core.Logger("OrgasmStart");
				orgasmState = 0;
				orgasmCooldownTime = core.currentTime + orgasmCooldownTimeOffset;
				orgasmUpdateTime = core.currentTime;
				core.customUi.jsOrgasmCount.val += 1.0f;
				if((int)core.customUi.jsOrgasmCount.val >= 5){
					core.customUi.jsHasMultipleOrgasms.val = true;
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void OrgasmFinish(){
			try{
				core.Logger("OrgasmFinish");
				orgasmCooldownTime = 0f;
				orgasmUpdateTime = core.currentTime;
				core.isDuringOrgasm = false;
				core.isDuringAfterglow = false;
				core.isBreathActive = false;
//				core.belly.restoreStatus();
//				core.belly.restoreRotation();
				orgasmState = -1;
				core.SetGauge(0);
				core.calmingDownPlus = 1;
				core.calmingDownTimeMinusOffset = -1;
				core.StageChangingCheck();
				core.worldUi.OnOrgasmStop();
				CalcGauge();
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void CalcGauge(){
			try{
				if(core.stageThreshold == null) core.InitializeStageThreshold();

				int maxCount = 10;
				float recoveryRatio;
				int tmporgasmCount = ((int)core.customUi.jsOrgasmCount.val > maxCount) ? maxCount : (int)core.customUi.jsOrgasmCount.val;
				
				recoveryRatio = (float)(maxCount - 1) / (float)maxCount * ((float)tmporgasmCount / (float)maxCount);
				if(core.customUi.dJSSlider.ContainsKey("Ease of Re-Climax"))
				{
					recoveryRatio *= core.customUi.dJSSlider["Ease of Re-Climax"].val;
				}
				else
				{
					Utility.Debug("UnknownError: core.customUi.dJSSlider[\"Ease of Re-Climax\"] is not found.");
				}
				int tmpGauge = core.stageThreshold[2] + (int)(recoveryRatio * (float)(core.stageThreshold[5] - core.stageThreshold[2]));
				core.SetGauge(tmpGauge);
				core.StageChangingCheck();				
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public string WhereAmI(){
			string result = "";
			return result;
		}
		
		public void StateWaitForStart(float delta){
			//orgasmState:0
			if(delta < orgasmStartInterval){
				StateWaitForStart(delta);
			}else{
				StateChange(100);
				//orgasmState = 9;
				orgasmUpdateTime = core.currentTime;
			}
		}
		
		public void StateChange(int state){
			core.Logger("stage change:" + orgasmState + "=>" + state);
			orgasmState = state;
		}
		
		public void StateSetupCommonWaves(float delta, string type){
			//orgasmState:100
			if(type == "orgasm"){
				core.Logger("*** orgasm start.");
				firstTermCount = 0;
				core.isDuringOrgasm = true;
				core.isDuringAfterglow = false;
				isBigWave = true;
				core.isFirstTerm = true;
				core.isFirstWave = true;
				commonWaveCycleBig = UnityEngine.Random.Range(core.orgasmWaveCycleBigMin, core.orgasmWaveCycleBigMax);
//				core.belly.changePositionAndRotationState();
				core.worldUi.OnOrgasmStart();
			}else{
				core.Logger("*** afterglow start.");
				orgasmAfterglowCount = 0;
				core.isDuringAfterglow = true;
				core.isEmbarassed = true;
				isBigWave = true;
				core.isFirstWave = true;
				commonWaveCycleBig = UnityEngine.Random.Range(core.orgasmAfterglowCycleBigMin, core.orgasmAfterglowCycleBigMax);
				orgasmCooldownTime = core.currentTime + orgasmCooldownTimeOffset;
			}
			orgasmUpdateTime = core.currentTime;
			StateChange(101);
		}
		
		public void StateStartCommonWaves(float delta){
			core.worldUi.UpdateStatus();
			//orgasmState:101
			int waveCycleMin = (core.isDuringAfterglow) ? core.orgasmAfterglowCycleSmallMin : core.orgasmWaveCycleSmallMin;
			int waveCycleMax = (core.isDuringAfterglow) ? core.orgasmAfterglowCycleSmallMax : core.orgasmWaveCycleSmallMax;
			int waveCycleNum = UnityEngine.Random.Range(waveCycleMin, waveCycleMax);
			
			orgasmTimeRandomOffset1 = orgasmTimeOffset1 * UnityEngine.Random.Range(0.0f, 0.3f);
			orgasmTimeRandomOffset2 = orgasmTimeOffset2 * UnityEngine.Random.Range(0.0f, 0.3f);
			orgasmTimeStart = 0f;
			orgasmTimePeak = orgasmTimeOffset1 + orgasmTimeRandomOffset1;
			orgasmTimeFinish = orgasmTimeOffset1 + orgasmTimeRandomOffset1 + orgasmTimeOffset2 + orgasmTimeRandomOffset2;
			core.UpdateAll();
			orgasmUpdateTime = core.currentTime;

			float randContinuousRatio = UnityEngine.Random.Range(0.0f, 1.0f);
			randContinuousRatio *= (core.isFirstTerm) ? 0.5f : 1.0f;
			bool isWaveContinuous = randContinuousRatio < continuousBigWaveRatio;

			if(isBigWave){
				commonWaveCycleSmall = (isWaveContinuous && !core.isDuringAfterglow) ? 0 : waveCycleNum;
			}
			
			SoundMove();
			MouthMove();
			core.headController.PickAction(bIgnoringFiringRatio:false);
			
			StateChange(102);
		}

		public void StateProcessCommonWaves(float delta){
			//orgasmState:102
			if(delta >= orgasmTimeStart && delta < orgasmTimePeak){
				BellyMove(true);
				if(!core.isDuringAfterglow){
					core.customUnityAssetItem.PeeRandom(core.currentTime);
				}
				
			}else if(delta >= orgasmTimePeak && delta < orgasmTimeFinish){
				BellyMove(false);
				
			}else{
//				core.bodyMorph.applyBodyMorph();
				StateChange(103);
			}
		}

		public void StateCheckCommonWaveEnd(float delta){
			//orgasmState:103
			if(commonWaveCycleBig == 0 && commonWaveCycleSmall == 0){
				if(!core.isDuringAfterglow){
					StateChange(100);
				}else{
					OrgasmFinish();
				}
			}else{
				if(commonWaveCycleBig > 0){
					float intervalMax = core.isDuringAfterglow ? core.orgasmAfterglowCycleIntervalBigMax : core.orgasmWaveCycleIntervalBigMax;
					commonWaveCycleInterval = intervalMax * UnityEngine.Random.Range(0.8f, 3.0f);
				}else{
					float intervalMax = core.isDuringAfterglow ? core.orgasmAfterglowCycleIntervalSmallMax : core.orgasmWaveCycleIntervalSmallMax;
					commonWaveCycleInterval = intervalMax * UnityEngine.Random.Range(0.5f, 1.5f);
				}
				core.Logger("OrgasmControl: afterglow/big/small/waveState = " + core.isDuringAfterglow + "/" + commonWaveCycleBig + "/" + commonWaveCycleSmall + "/" + orgasmState);
				orgasmUpdateTime = core.currentTime;
				StateChange(104);
			}
		}
		
		public void StateCheckCommonIntervalEnd(float delta){
			//orgasmState:104
			if(delta < commonWaveCycleInterval){
				core.bodyMorph.ShiveringRandom();
				//nop
			}else{
				if(commonWaveCycleSmall > 0){
					core.isFirstWave = false;
					isBigWave = false;
					commonWaveCycleSmall--;
				}else if(commonWaveCycleSmall == 0){
					firstTermCount++;
					if(firstTermCount >= firstTermCountThreshold){
						core.isFirstTerm = false;
					}
					isBigWave = true;
					commonWaveCycleBig--;
					orgasmUpdateTime = core.currentTime;
				}
				orgasmUpdateTime = core.currentTime;
				core.Logger("OrgasmControl: afterglow/big/small/waveState = " + core.isDuringAfterglow + "/" + commonWaveCycleBig + "/" + commonWaveCycleSmall + "/" + orgasmState);
				StateChange(101);
			}
		}
					
			
		public void BellyMove(bool isBump){
			float strengthMax = (core.isDuringAfterglow) ? core.afterglowStrengthMax : core.orgasmStrengthMax;
			float poseStrength = isBigWave ? UnityEngine.Random.Range(strengthMax * 0.5f, strengthMax * 2.0f) : UnityEngine.Random.Range(strengthMax * 0.5f, strengthMax);
			poseStrength *= core.isFirstWave ? 2.0f : 1.0f;				
			float sign = (isBump) ? 1f : -1f;
			
			core.belly.BellyBump(poseStrength * sign);
			core.bodyMorph.ShiveringBump(isBump);
		}

		public void MouthMove(){
			if(core.stageHeat == "veryHot" || core.stageHeat == "nymphomania"){
				core.expression2.OpenMouth(3.0f);
			}			
		}
		
		public void SoundMove(){
			if(!isBigWave || core.isDuringAfterglow){ return; }
			string type = (core.isFirstWave) ? "climax" : "moan";
			core.sound.PlaySound(type, core.currentTime);
		}
		
		
		public void OrgasmControl(float time){
			try{
				if(orgasmState < 0){
					return;
				}

				float delta = time - orgasmUpdateTime;
				
				if(orgasmState == -1){
					//nop
				}else if(orgasmState == 0){
					StateWaitForStart(delta);
				}else if(orgasmState == 100){
					if(!core.isDuringOrgasm){
						StateSetupCommonWaves(delta, "orgasm");
					}else{
						StateSetupCommonWaves(delta, "afterglow");						
					}
				}else if(orgasmState == 101){
					StateStartCommonWaves(delta);
				}else if(orgasmState == 102){
					StateProcessCommonWaves(delta);
				}else if(orgasmState == 103){
					StateCheckCommonWaveEnd(delta);
				}else if(orgasmState == 104){
					StateCheckCommonIntervalEnd(delta);					
				}
			}catch (Exception e){
				core.LoggerE(e);
			}
		}
		
		public void OnDestroy(){
			
		}
		
	}
}
